#!/usr/bin/python3

#	control++ : System configuration tool
#	Copyright (C) 2017-2020 Alexey Appolonov
#
#	This program is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program.  If not, see <http://www.gnu.org/licenses/>.

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

import stat
import os
from shutil      import rmtree
from collections import defaultdict

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Base defines

TESTED_EXE = 'control++'
CONF_DIR   = '/etc/control++'
UNIT_PERM  = 'permissions'
UNIT_ULIM  = 'ulimits'
UNIT_STRT  = 'scripts'
MODE_SETTING = {
	UNIT_PERM: [TESTED_EXE, UNIT_PERM],
	UNIT_ULIM: [TESTED_EXE, UNIT_ULIM],
	UNIT_STRT: [TESTED_EXE, UNIT_STRT],
	}
PERM_MODE_SETTING = MODE_SETTING[UNIT_PERM]
ULIM_MODE_SETTING = MODE_SETTING[UNIT_ULIM]
STRT_MODE_SETTING = MODE_SETTING[UNIT_STRT]
SUBMODE_DESC = {
	UNIT_PERM: os.path.join(CONF_DIR, UNIT_PERM),
	UNIT_ULIM: os.path.join(CONF_DIR, UNIT_ULIM),
	UNIT_STRT: os.path.join(CONF_DIR, UNIT_STRT),
	}
UNIT_PERM_DESC = SUBMODE_DESC[UNIT_PERM]
UNIT_ULIM_DESC = SUBMODE_DESC[UNIT_ULIM]
UNIT_STRT_DESC = SUBMODE_DESC[UNIT_STRT]
EXEC_BITS = (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Decorator function that catches the exceptions raised in a passed function

def CatchExc(func):
	def Wrapper(self=None):
		try:
			func(self)
		except Exception as e:
			print(str(e))
			return False
		return True
	return Wrapper

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Test dirrectory

# 	[test_dir]:
#		[base_dir]:
#			[d1]:
#				[d1]:
#					f1, f2
#				f1, f2
#			[d1]:
#				f1, f2
#			[d3]
#			f1, f2
#		[d1]:
#			f1
#		[d2]:
#			f1
#		f1

EXEC_DIR = os.path.join('/var/lib', TESTED_EXE + '/tests')
TEST_DIR = os.path.join('/run', TESTED_EXE)
BASE_DIR = os.path.join(TEST_DIR, 'base_dir')
INSIDE_FILES = [os.path.join(BASE_DIR, os.path.join(d, f) if f else d) \
	for d in ('d1', 'd1/d1', 'd2', '') for f in ('', 'f1', 'f2') if f or d] + \
	[os.path.join(BASE_DIR, 'd3')]
OUTSIDE_FILES = [os.path.join(TEST_DIR, os.path.join(d, f) if f else d) \
	for d in ('d1', 'd2', '') for f in ('', 'f1') if f or d]
TEST_FILES = INSIDE_FILES + OUTSIDE_FILES + [BASE_DIR]
REG_FILES = [f for f in TEST_FILES if f[-2] == 'f']
DIR_FILES = [f for f in TEST_FILES if f[-2] != 'f']
TEST_FILES_ODD  = [f for f in TEST_FILES if os.path.basename(f) != 'base_dir' \
	and int(f[-1]) % 2 == 1]
REG_FILES_ODD = [f for f in TEST_FILES_ODD if f in REG_FILES]
DIR_FILES_ODD = [f for f in TEST_FILES_ODD if f in DIR_FILES]

class TestDir():

	def __init__(self, recreate=False, printer=None):
		self.printer = printer
		self.recreate = recreate

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	def Prepare(self):

		# Creating the test dir itself
		if not self.__CreateDir(TEST_DIR):
			return False

		# Creating the files of the test file structure
		testdir_level = len(TEST_DIR.split('/'))
		for path in TEST_FILES:
			files = path.split('/')
			for level in range(testdir_level, len(files)):
				subpath = '/'.join(files[:level + 1])
				if os.path.exists(subpath):
					continue
				self.__Print(f'Creating "{subpath}"')
				if os.path.normpath(subpath)[-2] != 'f':
					os.mkdir(subpath)
				else:
					os.mknod(subpath, mode = 0o777)

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	def Clear(self):
		pycache_dir = os.path.join(EXEC_DIR, '__pycache__')
		return (
			self.__ClearDir(TEST_DIR) and
			self.__ClearDir(pycache_dir, rmdir=True)
			)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	def __ClearDir(self, dir_name, rmdir=False):

		if not os.path.isdir(dir_name):
			self.__Print(f'"{dir_name}" is not a directory')
			return False

		for file_name in os.listdir(dir_name):
			full_file_path = os.path.join(dir_name, file_name)
			try:
				if os.path.isfile(full_file_path):
					os.remove(full_file_path)
				elif os.path.isdir(full_file_path):
					rmtree(full_file_path)
			except Exception as e:
				self.__Print(str(e))
				return False

		if rmdir:
			try:
				os.rmdir(dir_name)
			except Exception as e:
				self.__Print(str(e))
				return False

		return True

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	def __Print(self, message):
		if self.printer:
			self.printer.LineEnd('\t' + message)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	def __CreateDir(self, path):

		if not path:
			self.__Print('Empty path')
			return False

		EXISTS = os.path.exists(path)
		IS_DIR = os.path.isdir(path)

		if EXISTS and not IS_DIR:
			self.__Print(f'Can\'t create "{path}" - there is a file '
				'with a same name')
			return False
		if EXISTS and IS_DIR:
			if self.recreate:
				if not self.__ClearDir(path):
					self.__Print('Can\'t recreate ' + path)
					return False
			else:
				self.__Print(path + ' is already exist')
				return True
		elif not EXISTS:
			self.__Print('Creating ' + path)
			os.makedirs(path)

		return True

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Mode description

class ModeDesc():

	def __init__(self, path):
		self.sections = []
		self.current_sec = ''
		self.path = path

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	def __Stringify(self):
		s = ''
		for sec in self.sections:
			for sec_name, fields in sec.items():
				s = f'[{sec_name}]\n' + '\n'.join(fields)
		return s

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	def NewSec(self, sec_name):
		self.sections.append(defaultdict(list))
		self.current_sec = sec_name

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	def AddField(self, field):
		if self.current_sec:
			self.sections[-1][self.current_sec].append(field)

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	@CatchExc
	def Write(self):
		with open(self.path, 'w') as f:
			f.write(self.__Stringify())

	# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

	@CatchExc
	def Clear(self):
		# Clearing every mode description used during testing
		lists = [os.path.join(UNIT_PERM_DESC, f'test_{l}list')
			for l in ('black', 'white')]
		for path in lists + [self.path]:
			if os.path.isfile(path):
				os.remove(path)
